@@ -1,4 +1,6 @@
-import os, subprocess, runpy
+import os
+import runpy
+import subprocess
 
 # Try to get and update the version.
 
@@ -9,16 +11,19 @@ try:
     if "HY_VERSION" in os.environ:
         __version__ = os.environ["HY_VERSION"]
     else:
-        __version__ = (subprocess.check_output
-                       (["git", "describe", "--tags", "--dirty"])
-                       .decode('ASCII').strip()
-                       .replace('-', '+', 1).replace('-', '.'))
+        __version__ = (
+            subprocess.check_output(["git", "describe", "--tags", "--dirty"])
+            .decode("ASCII")
+            .strip()
+            .replace("-", "+", 1)
+            .replace("-", ".")
+        )
 
     with open(VERSIONFILE, "wt") as o:
         o.write("__version__ = {!r}\n".format(__version__))
 
 except (subprocess.CalledProcessError, OSError):
     if os.path.exists(VERSIONFILE):
-        __version__ = runpy.run_path(VERSIONFILE)['__version__']
+        __version__ = runpy.run_path(VERSIONFILE)["__version__"]
     else:
         __version__ = "unknown"
