@@ -16,20 +16,17 @@ hy.importer._inject_builtins()
 # we import for side-effects.
 
 
-class M:
-    """``hy.M`` is an object that provides syntactic sugar for imports. It allows syntax like ``(hy.M.math.sqrt 2)``  to mean ``(import math) (math.sqrt 2)``, except without bringing ``math`` or ``math.sqrt`` into scope. This is useful in macros to avoid namespace pollution. To refer to a module with dots in its name, use slashes instead: ``hy.M.os/path.basename`` gets the function ``basename`` from the module ``os.path``.
+class I:
+    """``hy.I`` is an object that provides syntactic sugar for imports. It allows syntax like ``(hy.I.math.sqrt 2)``  to mean ``(import math) (math.sqrt 2)``, except without bringing ``math`` or ``math.sqrt`` into scope. (See :ref:`hy.R <hy.R>` for a version that requires a macro instead of importing a Python object.) This is useful in macros to avoid namespace pollution. To refer to a module with dots in its name, use slashes instead: ``hy.I.os/path.basename`` gets the function ``basename`` from the module ``os.path``.
 
-    You can also call ``hy.M`` like a function, as in ``(hy.M "math")``, which is useful when the module name isn't known until run-time. This interface just calls :py:func:`importlib.import_module`, avoiding (1) mangling due to attribute lookup, and (2) the translation of ``/`` to ``.`` in the module name. The advantage of ``(hy.M modname)`` over ``importlib.import_module(modname)`` is merely that it avoids bringing ``importlib`` itself into scope."""
+    You can also call ``hy.I`` like a function, as in ``(hy.I "math")``, which is useful when the module name isn't known until run-time. This interface just calls :py:func:`importlib.import_module`, avoiding (1) mangling due to attribute lookup, and (2) the translation of ``/`` to ``.`` in the module name. The advantage of ``(hy.I modname)`` over ``importlib.import_module(modname)`` is merely that it avoids bringing ``importlib`` itself into scope."""
     def __call__(self, module_name):
         import importlib
         return importlib.import_module(module_name)
     def __getattr__(self, s):
-        import re
-        return self(hy.mangle(re.sub(
-            r'/(-*)',
-            lambda m: '.' + '_' * len(m.group(1)),
-            hy.unmangle(s))))
-M = M()
+        from hy.reader.mangling import slashes2dots
+        return self(slashes2dots(s))
+I = I()
 
 
 # Import some names on demand so that the dependent modules don't have
@@ -40,7 +37,7 @@ _jit_imports = dict(
     read_many="hy.reader",
     mangle="hy.reader",
     unmangle="hy.reader",
-    eval=["hy.compiler", "hy_eval"],
+    eval=["hy.compiler", "hy_eval_user"],
     repr=["hy.core.hy_repr", "hy_repr"],
     repr_register=["hy.core.hy_repr", "hy_repr_register"],
     gensym="hy.core.util",
