@@ -5,13 +5,15 @@ import io
 import os
 import platform
 import py_compile
+import re
 import runpy
 import sys
+import types
 from contextlib import nullcontext
 from pathlib import Path
 
 import hy
-from hy._compat import PY3_9, PYPY
+from hy._compat import PY3_9
 from hy.compiler import hy_compile, hy_eval
 from hy.errors import HyLanguageError, filtered_hy_exceptions, hy_exc_handler
 from hy.importer import runhy
@@ -35,7 +37,7 @@ def run_command(source, filename=None):
     with filtered_hy_exceptions():
         try:
             hy_eval(
-                read_many(source, filename=filename, skip_shebang=True),
+                read_many(source, filename=filename),
                 __main__.__dict__,
                 __main__,
                 filename=filename,
@@ -47,28 +49,7 @@ def run_command(source, filename=None):
     return 0
 
 
-def run_icommand(source, **kwargs):
-    if Path(source).exists():
-        filename = source
-        set_path(source)
-        with open(source, "r", encoding="utf-8") as f:
-            source = f.read()
-    else:
-        filename = "<string>"
-
-    hr = REPL(**kwargs)
-    with filtered_hy_exceptions():
-        res = hr.runsource(source, filename=filename)
-
-    # If the command was prematurely ended, show an error (just like Python
-    # does).
-    if res:
-        hy_exc_handler(sys.last_type, sys.last_value, sys.last_traceback)
-
-    return hr.run()
-
-
-USAGE = "hy [-h | -v | -i CMD | -c CMD | -m MODULE | FILE | -] [ARG]..."
+USAGE = "hy [-h | -v | -i | -c CMD | -m MODULE | FILE | -] [ARG]..."
 VERSION = "hy " + hy.__version__
 EPILOG = """
 FILE
@@ -84,7 +65,7 @@ class HyArgError(Exception):
     pass
 
 
-def cmdline_handler(scriptname, argv):
+def cmdline_handler(argv):
     # We need to terminate interpretation of options after certain
     # options, such as `-c`. So, we can't use `argparse`.
 
@@ -112,9 +93,8 @@ def cmdline_handler(scriptname, argv):
         ),
         dict(
             name=["-i"],
-            dest="icommand",
-            terminate=True,
-            help="program passed in as string, then stay in REPL",
+            action="store_true",
+            help="launch REPL after running script; forces a prompt even if stdin does not appear to be a terminal",
         ),
         dict(
             name=["-m"],
@@ -240,43 +220,69 @@ def cmdline_handler(scriptname, argv):
         print(VERSION)
         return 0
 
-    if "command" in options:
-        sys.argv = ["-c"] + argv
-        return run_command(options["command"], filename="<string>")
+    action, action_arg = (
+        # If the `command` or `mod` options were provided, we'll run
+        # the corresponding code.
+        ["eval_string", options["command"]]
+            if "command" in options else
+        ["run_module", options["mod"]]
+            if "mod" in options else
+        # Otherwise, we'll run any provided filename as a script (or
+        # standard input, if the filename is "-").
+        ["run_script_stdin", None]
+            if argv and argv[0] == "-" else
+        ["run_script_file", argv[0]]
+            if argv else
+        # With none of those arguments, we'll launch the REPL (if
+        # standard input is a TTY) or run a script from standard input
+        # (otherwise).
+        ["just_repl", None]
+            if sys.stdin.isatty() else
+        ["run_script_stdin", None])
+    repl = (
+        REPL(
+            spy = options.get("spy"),
+            output_fn = options.get("repl_output_fn"))
+        if "i" in options or action == "just_repl"
+        else None)
+    source = ''
 
-    if "mod" in options:
+    if action == "eval_string":
+        sys.argv = ["-c"] + argv
+        if repl:
+            source = action_arg
+            filename = '<string>'
+        else:
+            return run_command(action_arg, filename="<string>")
+    elif action == "run_module":
+        if repl: raise ValueError()
         set_path("")
         sys.argv = [program] + argv
-        runpy.run_module(hy.mangle(options["mod"]), run_name="__main__", alter_sys=True)
+        runpy.run_module(hy.mangle(action_arg), run_name="__main__", alter_sys=True)
         return 0
-
-    if "icommand" in options:
-        return run_icommand(
-            options["icommand"],
-            spy=options.get("spy"),
-            output_fn=options.get("repl_output_fn"),
-        )
-
-    if argv:
-        if argv[0] == "-":
-            # Read the program from stdin
+    elif action == "run_script_stdin":
+        sys.argv = argv
+        if repl:
+            source = sys.stdin
+            filename = 'stdin'
+        else:
             return run_command(sys.stdin.read(), filename="<stdin>")
-
+    elif action == "run_script_file":
+        sys.argv = argv
+        filename = Path(action_arg)
+        set_path(filename)
+        # Ensure __file__ is set correctly in the code we're about
+        # to run.
+        if PY3_9:
+            if not filename.is_absolute():
+                filename = Path.cwd() / filename
+            if platform.system() == "Windows":
+                filename = os.path.normpath(filename)
+        if repl:
+            source = Path(filename).read_text()
+            repl.compile.compiler.skip_next_shebang = True
         else:
-            # User did "hy <filename>"
-
-            filename = Path(argv[0])
-            set_path(filename)
-            # Ensure __file__ is set correctly in the code we're about
-            # to run.
-            if PY3_9 and not PYPY:
-                if not filename.is_absolute():
-                    filename = Path.cwd() / filename
-                if platform.system() == "Windows":
-                    filename = os.path.normpath(filename)
-
             try:
-                sys.argv = argv
                 with filtered_hy_exceptions():
                     runhy.run_path(str(filename), run_name="__main__")
                 return 0
@@ -291,15 +297,33 @@ def cmdline_handler(scriptname, argv):
             except HyLanguageError:
                 hy_exc_handler(*sys.exc_info())
                 sys.exit(1)
+    else:
+        assert action == "just_repl"
 
-    return REPL(spy=options.get("spy"), output_fn=options.get("repl_output_fn")).run()
+    # If we didn't return earlier, we'll be using the REPL.
+    if source:
+        # Execute `source` in the REPL before entering interactive mode.
+        res = None
+        filename = str(filename)
+        with filtered_hy_exceptions():
+            accum = ''
+            for chunk in ([source] if isinstance(source, str) else source):
+                accum += chunk
+                res = repl.runsource(accum, filename=filename)
+                if not res:
+                    accum = ''
+        # If the command was prematurely ended, show an error (just like Python
+        # does).
+        if res:
+            hy_exc_handler(sys.last_type, sys.last_value, sys.last_traceback)
+    return repl.run()
 
 
 # entry point for cmd line script "hy"
 def hy_main():
     sys.path.insert(0, "")
     try:
-        sys.exit(cmdline_handler("hy", sys.argv))
+        sys.exit(cmdline_handler(sys.argv))
     except HyArgError as e:
         print(e)
         exit(1)
@@ -331,14 +355,18 @@ def hyc_main():
     return rv
 
 
-def hy2py_worker(source, options, filename, output_filepath=None):
+def hy2py_worker(source, options, filename=None, parent_module=None, output_filepath=None):
+    source_path = None
     if isinstance(source, Path):
+        source_path = source
         source = source.read_text(encoding="UTF-8")
+        if parent_module is None:
+            set_path(source_path)
 
     if not output_filepath and options.output:
         output_filepath = options.output
 
-    set_path(filename)
+
     with (
         open(output_filepath, "w", encoding="utf-8")
         if output_filepath
@@ -346,19 +374,33 @@ def hy2py_worker(source, options, filename, output_filepath=None):
     ) as output_file:
 
         def printing_source(hst):
-            for node in hst:
-                if options.with_source:
-                    print(node, file=output_file)
-                yield node
+            def _printing_gen(hst):
+                for node in hst:
+                    if options.with_source:
+                        print(node, file=output_file)
+                    yield node
+            printing_hst = hy.models.Lazy(_printing_gen(hst))
+            printing_hst.source = hst.source
+            printing_hst.filename = hst.filename
+            printing_hst.reader = hst.reader
+            return printing_hst
 
-        hst = hy.models.Lazy(
-            printing_source(read_many(source, filename, skip_shebang=True))
-        )
-        hst.source = source
-        hst.filename = filename
+        hst = printing_source(read_many(source, filename, skip_shebang=True))
 
         with filtered_hy_exceptions():
-            _ast = hy_compile(hst, "__main__", filename=filename, source=source)
+            module_name = source_path.stem if source_path else Path(filename).name
+            if parent_module:
+                module_name = f"{parent_module}.{module_name}"
+            module = types.ModuleType(module_name)
+            sys.modules[module_name] = module
+            try:
+                _ast = hy_compile(
+                     hst,
+                     module,
+                     filename=filename,
+                     source=source)
+            finally:
+                del sys.modules[module_name]
 
         if options.with_source:
             print()
@@ -377,17 +419,19 @@ def hy2py_worker(source, options, filename, output_filepath=None):
 def hy2py_main():
     options = dict(
         prog="hy2py",
-        usage="%(prog)s [options] [FILE]",
+        usage="%(prog)s [options] [-m MODULE | FILE | -]",
         formatter_class=argparse.RawDescriptionHelpFormatter,
     )
     parser = argparse.ArgumentParser(**options)
-    parser.add_argument(
+    gp = parser.add_argument_group().add_mutually_exclusive_group()
+    gp.add_argument("-m", dest="module", help="convert Hy module (or all files in module)")
+    gp.add_argument(
         "FILE",
         type=str,
         nargs="?",
-        help='Input Hy code (can be file or directory) (use STDIN if "-" or '
-        "not provided)",
+        help='convert Hy source file',
     )
+    gp.add_argument("-", dest="use_stdin", action="store_true", help="read Hy from stdin")
     parser.add_argument(
         "--with-source",
         "-s",
@@ -413,12 +457,17 @@ def hy2py_main():
 
     options = parser.parse_args(sys.argv[1:])
 
-    if options.FILE is None or options.FILE == "-":
+    if options.use_stdin or (options.FILE is None and options.module is None):
         sys.path.insert(0, "")
         filename = "<stdin>"
         hy2py_worker(sys.stdin.read(), options, filename)
-    else:
-        filename = options.FILE
+    elif options.module:
+        if options.module[:1] == ".":
+            raise ValueError(
+                "Relative module names not supported"
+            )
+        sys.path.insert(0, "")
+        filename = options.module.replace(".", os.sep)
         if os.path.isdir(filename):
             # handle recursively if --output is specified
             if not options.output:
@@ -426,7 +475,7 @@ def hy2py_main():
                     f"{filename} is a directory but the output directory is not specified. Use --output or -o in command line arguments to specify the output directory."
                 )
             os.makedirs(options.output, exist_ok=True)
-            for path, subdirs, files in os.walk(filename):
+            for path, _, files in os.walk(filename):
                 for name in files:
                     filename_raw, filename_ext = os.path.splitext(name)
                     if filename_ext == ".hy":
@@ -434,20 +483,23 @@ def hy2py_main():
                         # make sure to follow original file structure
                         subdirectory = os.path.relpath(path, filename)
                         output_directory_path = os.path.join(
-                            options.output if options.output else path, subdirectory
+                            options.output, subdirectory
                         )
                         os.makedirs(output_directory_path, exist_ok=True)
                         hy2py_worker(
                             Path(filepath),
                             options,
-                            filename,
+                            parent_module=path.replace(os.sep, "."),
                             output_filepath=os.path.join(
                                 output_directory_path, filename_raw + ".py"
                             ),
                         )
-
         else:
-            hy2py_worker(Path(options.FILE), options, filename)
+            filename += ".hy"
+            parent_module = ".".join(options.module.split(".")[:-1])
+            hy2py_worker(Path(filename), options, parent_module=parent_module)
+    else:
+        hy2py_worker(Path(options.FILE), options, options.FILE)
     parser.exit(0)
 
 
