@@ -34,7 +34,7 @@ class Object:
     `end_column` 3.
     """
 
-    properties = ["module", "_start_line", "_end_line", "_start_column", "_end_column"]
+    properties = ["_start_line", "_end_line", "_start_column", "_end_column"]
 
     def replace(self, other, recursive=False):
         if isinstance(other, Object):
@@ -203,7 +203,7 @@ class Symbol(Object, str):
     """
     Represents a symbol.
 
-    Symbol objects behave like strings under operations like :hy:func:`get`,
+    Symbol objects behave like strings under operations like :hy:func:`get <hy.pyops.get>`,
     :func:`len`, and :class:`bool`; in particular, ``(bool (hy.models.Symbol "False"))`` is true. Use :hy:func:`hy.eval` to evaluate a symbol.
     """
 
@@ -222,7 +222,7 @@ class Symbol(Object, str):
 
 
 _wrappers[bool] = lambda x: Symbol("True") if x else Symbol("False")
-_wrappers[type(None)] = lambda foo: Symbol("None")
+_wrappers[type(None)] = lambda _: Symbol("None")
 
 
 class Keyword(Object):
@@ -283,8 +283,8 @@ class Keyword(Object):
         :class:`hy.models.Keyword` objects).
 
         The optional second parameter is a default value; if provided, any
-        :class:`KeyError` from :hy:func:`get` will be caught, and the default returned
-        instead."""
+        :class:`KeyError` from :hy:func:`get <hy.pyops.get>` will be caught,
+        and the default returned instead."""
 
         from hy.reader import mangle
 
@@ -361,7 +361,7 @@ class Complex(Object, complex):
         if isinstance(real, str):
             value = super().__new__(cls, strip_digit_separators(real))
             p1, _, p2 = real.lstrip("+-").replace("-", "+").partition("+")
-            check_inf_nan_cap(p1, value.imag if "j" in p1 else value.real)
+            check_inf_nan_cap(p1, value.imag if "j" in p1.lower() else value.real)
             if p2:
                 check_inf_nan_cap(p2, value.imag)
             return value
@@ -537,7 +537,7 @@ class Dict(Sequence):
     Represents a literal :class:`dict`. ``keys``, ``values``, and ``items`` methods are
     provided, each returning a list, although this model type does none of the
     normalization of a real :class:`dict`. In the case of an odd number of child models,
-    ``keys`` returns the last child whereas ``values`` and ``items`` ignores it.
+    ``keys`` returns the last child whereas ``values`` and ``items`` ignore it.
     """
 
     def _pretty_str(self):
